# -*- coding: utf-8 -*-
# - Copyright (c) 2001-2016, IPython-Contrib Development Team
# - Copyright (c) 2016-, jupyter-contrib development team

"""Jupyter server extension to enable, disable and configure nbextensions."""

from __future__ import unicode_literals

import io
import json
import logging
import os.path
import posixpath
import re

import yaml
from notebook import version_info as nb_version_info
from notebook.base.handlers import APIHandler, IPythonHandler
from notebook.utils import url_path_join as ujoin
from notebook.utils import path2url
from tornado import web

# attempt to use LibYaml if available
try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

if nb_version_info < (5, 2, 0):
    from notebook.base.handlers import json_errors
else:
    # for notebook >= 5.2.0, instead of using json_errors, we must subclass
    # APIHandler. Since we already do this, just do nothing extra
    def json_errors(method):
        return method

__version__ = '0.4.1'

absolute_url_re = re.compile(r'^(f|ht)tps?://')


def _process_nbextension_spec(spec, relative_url_base=''):
    """
    Sanity-check and preprocess a spec loaded from a yaml descriptor file.

    Returns *either* a processed dict *or* a string error message describing
    why the spec was not suitable.
    """
    if not isinstance(spec, dict):
        return 'spec is not a dict, but an instance of {}'.format(type(spec))
    if 'Type' not in spec:
        return 'spec has no Type key'
    valid_types = {'IPython Notebook Extension', 'Jupyter Notebook Extension'}
    if str(spec['Type']).strip() not in valid_types:
        return 'spec has invalid value for Type key: {!r}'.format(spec['Type'])
    if 'Main' not in spec and 'require' not in spec:
        return 'spec has neither "Main" nor "require" key'
    # strip .js file extension from Main to give require path
    if 'require' not in spec:
        spec['require'] = os.path.splitext(spec['Main'])[0]

    spec.setdefault('Name', spec['require'])
    spec.setdefault('Compatibility', '?.x')
    spec.setdefault('Section', 'notebook')

    # generate relative URLs within the nbextensions namespace,
    # from urls relative to the yaml file
    for from_key, to_key in {
            'Link': 'readme', 'Icon': 'icon', 'Main': 'require'}.items():
        # check for the to_key first, use from_key as backup
        # str needed in python 3, otherwise it ends up bytes
        from_val = str(spec.get(to_key, ''))
        if not from_val:
            from_val = str(spec.get(from_key, ''))
        if not from_val:
            continue
        if absolute_url_re.match(from_val):
            spec[to_key] = from_val
        else:
            spec[to_key] = posixpath.normpath(
                ujoin(relative_url_base, from_val))
    return spec


def get_configurable_nbextensions(
        nbextension_dirs, exclude_dirs=('mathjax',), as_dict=False, log=None):
    """Build a list of configurable nbextensions based on YAML descriptor files.

    descriptor files must:
      - be located under one of nbextension_dirs
      - have the file extension '.yaml' or '.yml'
      - contain (at minimum) the following keys:
        - Type: must be 'IPython Notebook Extension' or
                'Jupyter Notebook Extension'
        - Main: relative url of the nbextension's main javascript file
    """
    extension_dict = {}

    # Traverse through nbextension subdirectories to find all yaml files
    # However, don't check directories twice. See
    #   github.com/Jupyter-contrib/jupyter_nbextensions_configurator/issues/25
    already_checked = set()
    for root_nbext_dir in nbextension_dirs:
        if root_nbext_dir in already_checked:
            continue
        else:
            already_checked.add(root_nbext_dir)
        if log:
            log.debug(
                'Looking for nbextension yaml descriptor files in {}'.format(
                    root_nbext_dir))
        for direct, dirs, files in os.walk(root_nbext_dir, followlinks=True):
            # filter to exclude directories
            dirs[:] = [d for d in dirs if d not in exclude_dirs]
            for filename in files:
                if os.path.splitext(filename)[1] not in ['.yml', '.yaml']:
                    continue
                yaml_path = os.path.join(direct, filename)
                yaml_relpath = os.path.relpath(yaml_path, root_nbext_dir)
                with io.open(yaml_path, 'r', encoding='utf-8') as stream:
                    try:
                        extension = yaml.load(stream, Loader=SafeLoader)
                    except yaml.YAMLError:
                        if log:
                            log.warning(
                                'Failed to load yaml file {}'.format(
                                    yaml_relpath))
                        continue
                extension = _process_nbextension_spec(
                    extension,
                    relative_url_base=path2url(os.path.dirname(yaml_relpath)))
                if not isinstance(extension, dict):
                    continue
                require = extension['require']

                if log:
                    if require in extension_dict:
                        msg = 'nbextension {!r} has duplicate listings'.format(
                            extension['require'])
                        msg += ' in both {!r} and {!r}'.format(
                            yaml_path, extension_dict[require]['yaml_path'])
                        log.warning(msg)
                        extension['duplicate'] = True
                    else:
                        log.debug('Found nbextension {!r} in {}'.format(
                            extension['Name'], yaml_relpath))

                extension_dict[require] = {
                    'yaml_path': yaml_path, 'extension': extension}
    if as_dict:
        return extension_dict
    return [val['extension'] for val in extension_dict.values()]


class ConfiguratorLogger(logging.LoggerAdapter):
    """Logging adapter to prepend the serverextension name to log messages."""

    def __init__(self, logger):
        super(ConfiguratorLogger, self).__init__(logger, {})

    def process(self, msg, kwargs):
        return '[{}] {}'.format(__name__, msg), kwargs


class NBExtensionHandlerJSON(APIHandler):
    """
    Returns a json list describing the configurable nbextensions.

    Based on part of notebook.services.config.handlers.ConfigHandler
    """

    @APIHandler.log.getter
    def log(self):
        return ConfiguratorLogger(super(NBExtensionHandlerJSON, self).log)

    @web.authenticated
    @json_errors
    def get(self):
        self.set_header("Content-Type", 'application/json')
        nbapp_webapp = self.application
        nbextension_dirs = nbapp_webapp.settings['nbextensions_path']
        extension_list = get_configurable_nbextensions(
            nbextension_dirs=nbextension_dirs, log=self.log)
        self.finish(json.dumps(extension_list))


class NBExtensionHandlerPage(IPythonHandler):
    """Renders the nbextension configuration interface."""

    @IPythonHandler.log.getter
    def log(self):
        return ConfiguratorLogger(super(NBExtensionHandlerPage, self).log)

    @web.authenticated
    def get(self):
        """Render the nbextension configuration interface."""
        self.finish(self.render_template(
            'nbextensions_configurator.html',
            page_title='Nbextensions Configuration',
            **self.application.settings
        ))


class RenderExtensionHandler(IPythonHandler):
    """Renders markdown files as pages."""

    @IPythonHandler.log.getter
    def log(self):
        return ConfiguratorLogger(super(RenderExtensionHandler, self).log)

    @web.authenticated
    def get(self, path):
        """Render given markdown file."""
        if not path.endswith('.md'):
            # for all non-markdown items, we redirect to the actual file
            return self.redirect(self.base_url + path)
        self.finish(self.render_template(
            'rendermd.html',
            md_url=path,
            page_title=path,
            **self.application.settings
        ))


def load_jupyter_server_extension(nbapp):
    """Load and initialise the server extension."""
    logger = ConfiguratorLogger(nbapp.log)
    logger.debug('Loading {}'.format(__version__))
    webapp = nbapp.web_app

    # ensure our template gets into search path
    templates_dir = os.path.join(os.path.dirname(__file__), 'templates')
    logger.debug('  Editing templates path to add {}'.format(templates_dir))
    rootloader = webapp.settings['jinja2_env'].loader
    for loader in getattr(rootloader, 'loaders', [rootloader]):
        if hasattr(loader, 'searchpath') and \
                templates_dir not in loader.searchpath:
            loader.searchpath.append(templates_dir)

    base_url = webapp.settings['base_url']

    # make sure our static files are available
    static_files_path = os.path.normpath(os.path.join(
        os.path.dirname(__file__), 'static'))
    logger.debug(
        '  Editing nbextensions path to add {}'.format(static_files_path))
    if static_files_path not in webapp.settings['nbextensions_path']:
        webapp.settings['nbextensions_path'].append(static_files_path)

    # add our new custom handlers
    logger.debug('  Adding new handlers')
    new_handlers = [(ujoin(base_url, '/nbextensions/' + u), h) for u, h in [
        (r"?", NBExtensionHandlerPage),
        (r"nbextensions_configurator/list$", NBExtensionHandlerJSON),
        (r"nbextensions_configurator/rendermd/(.*)", RenderExtensionHandler),
    ]]
    webapp.add_handlers(".*$", new_handlers)

    logger.info('enabled {}'.format(__version__))

def _jupyter_nbextension_paths():
    return [
        dict(
            section="notebook",
            src="static/nbextensions_configurator",
            dest="nbextensions_configurator",
            require='nbextensions_configurator/config_menu/main',
        ),
        dict(
            section="tree",
            src="static/nbextensions_configurator",
            dest="nbextensions_configurator",
            require='nbextensions_configurator/tree_tab/main',
        ),
    ]

def _jupyter_server_extension_paths():
    return [{
        'module': __name__
    }]



# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import logging
import os
import pipes
import sys
import threading
import time
from subprocess import PIPE, STDOUT, Popen

from jupyter_contrib_core.testing_utils import get_logger
from nose.plugins.skip import SkipTest
from selenium.webdriver.support.ui import WebDriverWait
from tornado import gen
from tornado.ioloop import IOLoop
from traitlets import default
from traitlets.config.application import LevelFormatter

from nbextensions_test_base import get_wrapped_logger, wrap_logger_handlers
from test_nbextensions_configurator import ConfiguratorTest

# run some basic checks to see if jupyterhub is supported on this platform
if os.name in ('nt', 'dos'):
    raise SkipTest('jupyterhub is not supported on Windows.')
elif sys.version_info[:2] < (3, 3):
    raise SkipTest('jupyterhub requires Python version 3.3 or above.')
else:
    try:
        import jupyterhub  # noqa
    except ImportError:
        raise SkipTest(
            'could not import jupyterhub, so skipping jupyterhub tests')
    from jupyterhub.auth import Authenticator
    from jupyterhub.spawner import LocalProcessSpawner
    from jupyterhub.tests.mocking import MockHub, public_url
    from jupyterhub.utils import random_port


class TestSpawner(LocalProcessSpawner):
    """Mock spawner, skipping user-switching that'd need root permissions."""

    @default('log')
    def _log_default(self):
        """wrap loggers for this application."""
        return wrap_logger_handlers(LocalProcessSpawner._log_default(self))

    def make_preexec_fn(self, *a, **kw):
        # skip the setuid stuff
        return

    def user_env(self, env):
        # copy select environment variables.
        env['USER'] = self.user.name
        env.update({k: v for k, v in dict(os.environ).items() if k in [
            'JUPYTER_CONFIG_DIR', 'JUPYTER_DATA_DIR', 'JUPYTER_RUNTIME_DIR',
            'HOME', 'SHELL',
        ]})
        return env

    @gen.coroutine
    def start(self):
        """Start the process. Overridden in order to capture output."""
        self.port = random_port()

        env = self.get_env()
        cmd = []
        cmd.extend(self.cmd)
        cmd.extend(self.get_args())

        self.log.info("Spawning %s", ' '.join(pipes.quote(s) for s in cmd))
        self.proc = Popen(
            cmd, env=env, preexec_fn=self.make_preexec_fn(self.user.name),
            start_new_session=True,  # don't forward signals
            stdout=PIPE, stderr=STDOUT,
        )
        self.pid = self.proc.pid
        self._read_proc_stderr_thread = thrd = threading.Thread(
            target=self._read_proc_stderr, name='_read_proc_stderr')
        thrd.daemon = True
        thrd.start()
        return (self.ip or '127.0.0.1', self.port)

    def _read_proc_stderr(self):
        logr = get_logger(self.user.name)
        logr.handlers[0].setFormatter(logging.Formatter(fmt='    %(message)s'))
        logr = wrap_logger_handlers(logr)
        for line in iter(self.proc.stdout.readline, b''):
            logr.info(line.decode('utf-8').strip('\n'))


TestSpawner.debug.default_value = True


class MockAuthenticator(Authenticator):
    """
    Dummy authentication. Returns the username if login is successful.

    Returns None otherwise.
    """

    _default_whitelist = {'nandy', 'aela'}

    @default('log')
    def _log_default(self):
        """wrap loggers for this application."""
        return wrap_logger_handlers(Authenticator._log_default(self))

    @gen.coroutine
    def authenticate(self, handler, data):
        username, password = data['username'], data['password']
        # just use equality for testing
        return (username if password == username else None)

    @default('whitelist')
    def get_default_whitelist(self):
        return self._default_whitelist


class JupyterHubConfiguratorTest(ConfiguratorTest):
    """Base class for nbextensions test cases running through jupyterhub."""

    uses_jupyterhub = True  # attribute used for filtering nose tests

    # we can't patch jupyter paths for single-user server running in a
    # subprocess, but we can patch environment variables passed to the
    # subprocess, so do a user install, to take advantage of the $HOME variable
    _install_user = True

    @classmethod
    def base_url(cls):
        return public_url(cls.app, cls.user)

    @classmethod
    def setup_class(cls):
        cls._failure_occurred = False  # flag for logging
        cls.log = get_wrapped_logger(cls.__name__)
        cls.log.handlers[0].setFormatter(LevelFormatter(
            fmt=(
                '[%(levelname)1.1s '
                '%(asctime)s.%(msecs).03d '
                '%(name)s %(module)s:%(lineno)d]'
                '%(message)s'
            ),
            datefmt='%H:%M:%S',
        ))
        cls._setup_patches()
        cls.pre_server_setup()

        cls.log.info('starting webdriver')
        cls.init_webdriver()
        cls.log.info('Starting jupyterhub server app thread')
        cls.app = MockHub.instance(
            log_datefmt="%H:%M:%S",
            authenticator_class=MockAuthenticator,
            spawner_class=TestSpawner,
        )
        # need to start jupyterhub app before calling super, as the super will
        # wait for the page to load
        try:
            cls.app.log = wrap_logger_handlers(cls.app.log)
            cls.app.start([])
        except Exception:
            cls._server_cleanup(
                error_msg='failed to start jupyterhub app')
            raise

        try:
            cls.log.info(
                'Logging into hub-spawned single-user notebook server.')
            login_url = public_url(cls.app) + 'login'
            cls.driver.get(login_url)
            cls.uname = name = next(iter(MockAuthenticator._default_whitelist))
            cls.wait_for_selector('#username_input').send_keys(name)
            cls.wait_for_selector('#password_input').send_keys(name)
            # This short wait seems necessary to avoid http 503 error
            time.sleep(1)
            cls.wait_for_selector('#login_submit').click()
            # wait for redirect to single-user server
            WebDriverWait(cls.driver, 30).until(
                lambda driver:
                    cls.app.users[name] is not None and
                    cls.app.users[name].server is not None and
                    cls.driver.current_url.startswith(
                        public_url(cls.app, cls.app.users[name])))
            # wait till single-user page loaded
            cls.wait_for_selector('#tab_content', timeout=10)

            user = cls.user = cls.app.users[name]
            if not user.running:
                io_loop = IOLoop()
                io_loop.make_current()
                io_loop.run_sync(user.spawn)
        except Exception:
            cls._server_cleanup(
                error_msg='failed to start/login to single-user server')
            raise

    @classmethod
    def _server_cleanup(cls, error_msg=None):
        if error_msg is not None:
            cls._failure_occurred = True
            cls.log.error(error_msg)
        cls.app.stop()
        # do cleanup explicitly as it's only registered using atexit
        cls.app.cleanup()

    @classmethod
    def teardown_class(cls):
        try:
            cls._server_cleanup()
            cls.app.__class__.clear_instance()
        finally:
            try:
                for ptch in cls.jupyter_patches:
                    ptch.stop()
            finally:
                try:
                    for func in cls.removal_funcs:
                        func()
                finally:
                    cls._print_logs_on_failure()



# -*- coding: utf-8 -*-
"""Base TestCase classes for nbextensions tests."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import logging
import os
from threading import Event, Thread

from jupyter_contrib_core.notebook_compat import serverextensions
from jupyter_contrib_core.testing_utils import (
    GlobalMemoryHandler, get_wrapped_logger, wrap_logger_handlers,
)
from jupyter_contrib_core.testing_utils.jupyter_env import patch_jupyter_dirs
from nose.plugins.skip import SkipTest
from notebook.notebookapp import NotebookApp
from notebook.tests.launchnotebook import NotebookTestBase
from tornado.ioloop import IOLoop
from traitlets.config import Config
from traitlets.traitlets import default

try:
    from unittest.mock import Mock
except ImportError:
    from mock import Mock  # py2

no_selenium = True
try:
    from selenium import webdriver
except ImportError:
    pass
else:
    no_selenium = False
    from selenium.common.exceptions import TimeoutException
    from selenium.webdriver.common.by import By
    from selenium.webdriver.remote import remote_connection
    from selenium.webdriver.support import expected_conditions as ec
    from selenium.webdriver.support.ui import WebDriverWait
    # don't show selenium debug logs
    remote_connection.LOGGER.setLevel(logging.INFO)


class NoseyNotebookApp(NotebookApp):
    """Wrap the regular logging handler(s). For use inside nose tests."""

    @default('log')
    def _log_default(self):
        """wrap loggers for this application."""
        return wrap_logger_handlers(NotebookApp._log_default(self))


class NbextensionTestBase(NotebookTestBase):
    """
    Base class for nbextensions test case classes.

    We override the setup_class method from NotebookTestBase in order to
    install things, and also to set log_level to debug.
    Also split some of the setup_class method into separate methods in order to
    simplify subclassing.
    """
    config = Config(NotebookApp={'log_level': logging.DEBUG})

    # these are added for notebook < 4.1, where url_prefix wasn't defined.
    # However, due to the fact that the base_url body data attribute in the
    # page template isn't passed through the urlencode jinja2 filter,
    # we can't expect a base_url which would need encoding to work :(
    if not hasattr(NotebookTestBase, 'url_prefix'):
        url_prefix = '/ab/'

        @classmethod
        def base_url(cls):
            return 'http://localhost:%i%s' % (cls.port, cls.url_prefix)

    _install_user = False
    _install_sys_prefix = False

    @classmethod
    def pre_server_setup(cls):
        """Setup extensions etc before running the notebook server."""
        # added to install things!
        cls.log.info('Enabling jupyter_nbextensions_configurator')
        inst_func = serverextensions.toggle_serverextension_python
        inst_funcname = '.'.join([inst_func.__module__, inst_func.__name__])
        logger = get_wrapped_logger(
            name=inst_funcname, log_level=logging.DEBUG)
        serverextensions.toggle_serverextension_python(
            'jupyter_nbextensions_configurator', enabled=True, logger=logger,
            user=cls._install_user, sys_prefix=cls._install_sys_prefix)

    @classmethod
    def get_server_kwargs(cls, **overrides):
        kwargs = dict(
            port=cls.port,
            port_retries=0,
            open_browser=False,
            runtime_dir=cls.jupyter_dirs['server']['runtime'],
            notebook_dir=cls.jupyter_dirs['server']['notebook'],
            base_url=cls.url_prefix,
            config=cls.config,
        )
        # disable auth-by-default, introduced in notebook PR #1831
        if 'token' in NotebookApp.class_trait_names():
            kwargs['token'] = ''
        kwargs.update(overrides)
        return kwargs

    @classmethod
    def start_server_thread(cls, started_event):
        """
        Start a notebook server in a separate thread.

        The start is signalled using the passed Event instance.
        """
        cls.log.info('Starting notebook server app thread')
        app = cls.notebook = NoseyNotebookApp(**cls.get_server_kwargs())
        # don't register signal handler during tests
        app.init_signal = lambda: None
        app.initialize(argv=[])
        loop = IOLoop.current()
        loop.add_callback(started_event.set)
        try:
            app.start()
        finally:
            # set the event, so failure to start doesn't cause a hang
            started_event.set()
            # app.session_manager.close call was added after notebook 4.0
            if hasattr(app.session_manager, 'close'):
                app.session_manager.close()

    @classmethod
    def _setup_patches(cls):
        (cls.jupyter_patches, cls.jupyter_dirs,
         remove_jupyter_dirs) = patch_jupyter_dirs()
        # store in a list to avoid confusion over bound/unbound method in pypy
        cls.removal_funcs = [remove_jupyter_dirs]
        try:
            for ptch in cls.jupyter_patches:
                ptch.start()

            # patches for items called in NotebookTestBase.teardown_class
            # env_patch needs a start method as well because of a typo in
            # notebook 4.0 which calls it in the teardown_class method
            cls.env_patch = cls.path_patch = Mock(['start', 'stop'])
            cls.home_dir = cls.config_dir = cls.data_dir = Mock(['cleanup'])
            cls.runtime_dir = cls.notebook_dir = Mock(['cleanup'])
            cls.tmp_dir = Mock(['cleanup'])
        except Exception:
            for func in cls.removal_funcs:
                func()
            raise

    @classmethod
    def setup_class(cls):
        """Install things & setup a notebook server in a separate thread."""
        cls.log = get_wrapped_logger(cls.__name__)
        cls._setup_patches()
        cls.pre_server_setup()
        try:
            started = Event()
            cls.notebook_thread = Thread(
                target=cls.start_server_thread, args=[started])
            cls.notebook_thread.start()
            started.wait()
            cls.wait_until_alive()
        except Exception:
            for func in cls.removal_funcs:
                func()
            raise

    @classmethod
    def teardown_class(cls):
        try:
            # call superclass to stop notebook server
            super(NbextensionTestBase, cls).teardown_class()
        finally:
            try:
                for ptch in cls.jupyter_patches:
                    ptch.stop()
            finally:
                for func in cls.removal_funcs:
                    func()


def _skip_if_no_selenium():
    if no_selenium:
        raise SkipTest('Selenium not installed. '
                       'Skipping selenium-based test.')
    if os.environ.get('TRAVIS_OS_NAME') == 'osx':
        raise SkipTest("Don't do selenium tests on travis osx")


class SeleniumNbextensionTestBase(NbextensionTestBase):

    # browser logs from selenium aren't very useful currently, but if you want
    # them, you can set the class attribute show_driver_logs to have them
    # output via the GlobalMemoryHandler on test failure
    show_driver_logs = False

    @classmethod
    def setup_class(cls):
        cls.init_webdriver()
        cls._failure_occurred = False  # flag for logging
        super(SeleniumNbextensionTestBase, cls).setup_class()

    @classmethod
    def init_webdriver(cls):
        cls.log = get_wrapped_logger(cls.__name__)
        _skip_if_no_selenium()

        if hasattr(cls, 'driver'):
            return cls.driver
        if (os.environ.get('CI') and os.environ.get('TRAVIS') and
                os.environ.get('SAUCE_ACCESS_KEY')):
            cls.log.info(
                'Running in CI environment. Using Sauce remote webdriver.')
            username = os.environ['SAUCE_USERNAME']
            access_key = os.environ['SAUCE_ACCESS_KEY']
            capabilities = {
                # 'platform': 'Mac OS X 10.9',
                'platform': 'Linux',
                'browserName': 'firefox',
                'version': 'latest',
                'tags': [os.environ['TOXENV'], 'CI'],
                'name': cls.__name__
            }
            hub_url = 'http://{}:{}@ondemand.saucelabs.com:80/wd/hub'.format(
                username, access_key)
            if os.environ.get('TRAVIS'):
                # see https://docs.travis-ci.com/user/gui-and-headless-browsers
                # and https://docs.travis-ci.com/user/sauce-connect
                capabilities.update({
                    'tunnel-identifier': os.environ['TRAVIS_JOB_NUMBER'],
                    'build': os.environ['TRAVIS_BUILD_NUMBER'],
                })
            cls.driver = webdriver.Remote(
                desired_capabilities=capabilities, command_executor=hub_url)
        else:
            cls.log.info('Using local webdriver.')
            cls.driver = webdriver.Firefox()
        return cls.driver

    def run(self, results):
        """Run a given test. Overridden in order to access results."""
        # in py2 unittest, run doesn't return the results object, so we need to
        # create one in order to have a reference to it.
        if results is None:
            results = self.defaultTestResult()
        super(SeleniumNbextensionTestBase, self).run(results)
        if results.failures or results.errors:
            self.__class__._failure_occurred = True
        return results

    @classmethod
    def _print_logs_on_failure(cls):
        if cls._failure_occurred:
            cls.log.info('\n'.join([
                '',
                '\t\tFailed test!',
                '\t\tCaptured logging:',
            ]))
            GlobalMemoryHandler.rotate_buffer(1)
            GlobalMemoryHandler.flush_to_target()

            browser_logger = get_wrapped_logger(
                name=cls.__name__ + '.driver', log_level=logging.DEBUG)
            if cls.show_driver_logs:
                cls.log.info('\n\t\tjavascript console logs below...\n\n')
                for entry in cls.driver.get_log('browser'):
                    level = logging._nameToLevel.get(
                        entry['level'], logging.ERROR)
                    msg = entry['message'].strip()
                    browser_logger.log(level, msg)
                    record, target = GlobalMemoryHandler._buffer[-1]
                    record.ct = entry['timestamp'] / 1000.
                    GlobalMemoryHandler._buffer[-1] = record, target
                GlobalMemoryHandler.flush_to_target()

        if (not cls._failure_occurred) or os.environ.get('CI'):
            cls.log.info('closing webdriver')
            cls.driver.quit()
        else:
            cls.log.info('keeping webdriver open')

    @classmethod
    def teardown_class(cls):
        cls._print_logs_on_failure()
        super(SeleniumNbextensionTestBase, cls).teardown_class()

    @classmethod
    def wait_for_element(cls, presence_cond, message, timeout=5):
        """WebDriverWait for an element to appear, fail test on timeout."""
        try:
            return WebDriverWait(cls.driver, timeout).until(
                ec.presence_of_element_located(presence_cond))
        except TimeoutException:
            if message:
                raise cls.failureException(message)
            else:
                raise cls.failureException(
                    '{}No element matching condition {!r} found in {}s'.format(
                        message, presence_cond, timeout))

    @classmethod
    def wait_for_selector(cls, css_selector, message='', timeout=5):
        """WebDriverWait for a selector to appear, fail test on timeout."""
        if message:
            message += '\n'
        message = '{}No element matching selector {!r} found in {}s'.format(
            message, css_selector, timeout)
        return cls.wait_for_element(
            (By.CSS_SELECTOR, css_selector), message=message, timeout=timeout)

    @classmethod
    def wait_for_partial_link_text(cls, link_text, message='', timeout=5):
        """WebDriverWait for a link to appear, fail test on timeout."""
        if message:
            message += '\n'
        message = (
            '{}No element matching partial link text '
            '{!r} found in {}s').format(message, link_text, timeout)
        return cls.wait_for_element((By.PARTIAL_LINK_TEXT, link_text),
                                    message=message, timeout=timeout)

    @classmethod
    def wait_for_xpath(cls, xpath, message='', timeout=5):
        """WebDriverWait for a selector to appear, fail test on timeout."""
        if message:
            message += '\n'
        message = '{}No element matching xpath {!r} found in {}s'.format(
            message, xpath, timeout)
        return cls.wait_for_element(
            (By.XPATH, xpath), message=message, timeout=timeout)



# -*- coding: utf-8 -*-

"""Test the package"""

from jupyter_contrib_core.testing_utils import raise_on_bad_version

from jupyter_nbextensions_configurator import __version__


def test_current_version():
    """check that version string complies with pep440"""
    raise_on_bad_version(__version__)



# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import io
import os
import random
import shutil
import time

import nose.tools as nt
import yaml
from jupyter_contrib_core.notebook_compat.nbextensions import _get_config_dir
from nose.plugins.skip import SkipTest
from notebook.services.config import ConfigManager
from notebook.utils import url_path_join
from selenium.common.exceptions import NoSuchElementException

import jupyter_nbextensions_configurator
from nbextensions_test_base import SeleniumNbextensionTestBase

# from http://nose.readthedocs.io/en/latest/writing_tests.html#writing-tests
#
# > nose runs functional tests in the order in which they appear in the module
# > file. TestCase-derived tests and other test classes are run in alphabetical
# > order
#
# So, I name test methods of classes using numbers to control execution order.


class ConfiguratorTest(SeleniumNbextensionTestBase):
    """Tests for the nbextensions_configurator server extension."""

    @property
    def nbext_configurator_url(self):
        return url_path_join(self.base_url(), 'nbextensions')

    @classmethod
    def pre_server_setup(cls):
        """Setup a temporary environment in which to run a notebook server."""
        super(ConfiguratorTest, cls).pre_server_setup()
        cls.add_test_yaml_files()

    def test_00_load_nbextensions_page(self):
        """Check that <base_url>/nbextensions url loads correctly."""
        self.driver.get(self.nbext_configurator_url)

    def test_01_page_title(self):
        nt.assert_in('extension', self.driver.title.lower())
        nt.assert_in('configuration', self.driver.title.lower())

    def test_02_body_data_attribute(self):
        nbext_list = self.driver.execute_script('return window.extension_list')
        # we no longer embed the list into the page
        nt.assert_is_none(nbext_list)

    def test_03_extension_ui_presence(self):
        self.wait_for_selector(
            '.nbext-ext-row', 'an nbextension ui should load')
        enable_links = self.driver.find_elements_by_css_selector(
            '.nbext-selector')
        nt.assert_greater(
            len(enable_links), 0, 'some nbextensions should have enable links')

    def test_04_readme_rendering(self):
        # load an nbextension UI whose readme contains an image to render
        self.wait_for_partial_link_text('dashboard').click()
        self.wait_for_selector('.nbext-readme > .panel-body img',
                               'there should be an image in the readme')

    def test_05_click_page_readme_link(self):
        self.driver.find_element_by_css_selector('.nbext-page-title a').click()
        self.wait_for_selector('#render-container img',
                               'there should be an image in the readme')

    def test_06_enable_tree_tab(self):
        self.driver.get(self.nbext_configurator_url)
        self.wait_for_selector(
            '.nbext-ext-row', 'an nbextension ui should load')
        # now enable the appropriate nbextension
        self.wait_for_partial_link_text(
            'dashboard'
        ).find_element_by_css_selector('.nbext-enable-toggle').click()
        self.check_extension_enabled(
            'tree', 'nbextensions_configurator/tree_tab/main',
            expected_status=True)

    def test_07_open_tree_tab(self):
        self.driver.get(self.base_url())
        self.wait_for_selector(
            '#tabs a[href$=nbextensions_configurator]').click()
        self.wait_for_selector(
            '.nbext-ext-row', 'an nbextension ui should load')

    def test_08_disable_tree_tab(self):
        # now disable the appropriate nbextension & wait for update to config
        self.wait_for_partial_link_text(
            'dashboard'
        ).find_element_by_css_selector('.nbext-enable-toggle').click()
        self.check_extension_enabled(
            'tree', 'nbextensions_configurator/tree_tab/main',
            expected_status=False)
        self.driver.get(self.base_url())
        with nt.assert_raises(AssertionError):
            self.wait_for_selector('#tabs a[href$=nbextensions_configurator]')

    def test_09_no_unconfigurable_yet(self):
        self.driver.get(self.nbext_configurator_url)
        self.wait_for_selector(
            '.nbext-ext-row', 'an nbextension ui should load')
        selector = self.driver.find_element_by_css_selector('.nbext-selector')
        nt.assert_not_in(
            'daemon', selector.text,
            'There should be no daemons in the selector yet')

    def test_10_refresh_list(self):
        # 'enable' a fake nbextension
        section, require = 'notebook', 'balrog/daemon'
        self.set_extension_enabled(section, require, True)
        # refresh the list to check that it appears
        self.wait_for_selector('.nbext-button-refreshlist').click()
        self.wait_for_partial_link_text(require)
        selector = self.driver.find_element_by_css_selector('.nbext-selector')
        nt.assert_in(
            'daemon', selector.text,
            'There should now be a daemon in the selector')

    def test_11_allow_configuring_incompatibles(self):
        require = 'balrog/daemon'
        # allow configuring incompatibles
        self.wait_for_selector(
            '#nbext_hide_incompat',
            'potentially incompatible nbextensions should show checkbox'
        ).click()
        # select it, now it's configurable
        self.driver.find_element_by_partial_link_text(require).click()

    def test_12_unconfigurable(self):
        section, require = 'notebook', 'balrog/daemon'
        sel_disable = '.nbext-enable-btns .btn:nth-child(2)'
        sel_forget = '.nbext-enable-btns .btn:nth-child(3)'
        # wait for ui to load
        self.wait_for_xpath('//h3[contains(text(), "daemon")]')
        # wait a second for the other nbextension ui to hide
        time.sleep(1)
        # there should be no forget button visible yet
        with nt.assert_raises(NoSuchElementException):
            self.driver.find_element_by_css_selector(sel_forget)
        # disable balrog
        self.wait_for_selector(sel_disable)
        visible_disablers = [
            el for el in self.driver.find_elements_by_css_selector(sel_disable)
            if el.is_displayed()]
        nt.assert_equal(1, len(visible_disablers),
                        'Only one disable button should be visible')
        visible_disablers[0].click()
        # now forget it
        self.wait_for_selector(sel_forget, 'A forget button should display')
        self.driver.find_element_by_css_selector(sel_forget).click()
        # confirm dialog
        self.wait_for_selector(
            '.modal-dialog .modal-footer .btn-danger',
            'a confirmation dialog should show'
        ).click()
        # it should no longer be enabled in config
        time.sleep(1)
        conf = self.get_config_manager().get(section)
        stat = conf.get('load_extensions', {}).get(require)
        nt.assert_is_none(
            stat, '{} should not have a load_extensions entry'.format(require))
        # and should no longer show in the list
        self.wait_for_selector(
            '.nbext-selector nav ul li', 'some nbextensions should show')
        nbext_sel = self.driver.find_element_by_css_selector('.nbext-selector')
        nt.assert_not_in(
            'daemon', nbext_sel.text,
            'There should no longer be a daemon in the selector')

    def test_13_duplicate_paths(self):
        if getattr(self, 'notebook', None) is None:
            raise SkipTest('cannot edit notebook server nbextensions path')

        # duplicate the dodgy/test entry on path
        saved = list(self.notebook.web_app.settings['nbextensions_path'])
        nt.assert_in(self.jupyter_dirs['dodgy']['nbexts'], saved)
        self.notebook.web_app.settings['nbextensions_path'].append(
            self.jupyter_dirs['dodgy']['nbexts'])
        try:
            self.driver.get(self.nbext_configurator_url)
            self.wait_for_selector('.nbext-selector')
            self.wait_for_partial_link_text('dummy').click()
            dummy = self.wait_for_xpath('''
//h3[contains(text(), "dummy")]
//ancestor::div[
    contains(concat(" ", normalize-space(@class), " "), " nbext-ext-row ")
]''')
            with nt.assert_raises(NoSuchElementException):
                dummy.find_element_by_css_selector('.alert-warning')
            with nt.assert_raises(NoSuchElementException):
                dummy.find_element_by_xpath(
                    './*[contains(text(),"different yaml files")]')
        finally:
            self.notebook.web_app.settings['nbextensions_path'] = saved

    @classmethod
    def get_config_manager(cls):
        try:
            # single-user notebook server tests use cls.notebook for app
            return cls.notebook.config_manager
        except AttributeError:
            # jupyterhub-based tests don't (can't) have cls.notebook defined,
            # so we must construct a ConfigManager from scratch
            return ConfigManager(
                log=cls.log,
                config_dir=os.path.join(_get_config_dir(user=True), 'nbconfig')
            )

    @classmethod
    def set_extension_enabled(cls, section, require, enabled):
        cm = cls.get_config_manager()
        new_conf = {}
        if enabled is not None:
            enabled = bool(enabled)
        new_conf.setdefault('load_extensions', {})[require] = enabled
        cm.update(section, new_conf)

    @classmethod
    def check_extension_enabled(cls, section, require, expected_status=True,
                                timeout=10, check_period=0.5):
        cm = cls.get_config_manager()
        for ii in range(0, max(1, int(timeout / check_period))):
            load_exts = cm.get(section).get('load_extensions', {})
            enabled = [req for req, en in load_exts.items() if en]
            if (require in enabled) == expected_status:
                break
            time.sleep(check_period)
        assert_func = (
            nt.assert_in if expected_status else nt.assert_not_in)
        assert_func(require, enabled,
                    'nbxtension should {}be in enabled list'.format(
                        '' if expected_status else 'not '))

    @classmethod
    def add_test_yaml_files(cls):
        """Add in dodgy yaml files in an extra nbextensions dir."""
        cls.jupyter_dirs['dodgy'] = {
            'nbexts': os.path.join(cls.jupyter_dirs['root'], 'dodgy', 'nbext')}
        dodgy_nbext_dir_path = cls.jupyter_dirs['dodgy']['nbexts']
        if not os.path.exists(dodgy_nbext_dir_path):
            os.makedirs(dodgy_nbext_dir_path)
        cls.config.NotebookApp.setdefault(
            'extra_nbextensions_path', []).append(dodgy_nbext_dir_path)

        # an invalid yaml file
        yaml_path_invalid = os.path.join(
            dodgy_nbext_dir_path, 'nbext_invalid_yaml.yaml')
        with io.open(yaml_path_invalid, 'w') as f:
            f.write('not valid yaml!: [')

        # various test yaml files
        test_yamls = {
            # we use str() because otherwise python2 will write
            # !!python/unicode in the yaml, and pyyaml SafeLoader will refuse
            # to reconstruct the objects :( boo!
            'not_an_nbext': [
                str('valid yaml'), str("doesn't always"),
                str('make for a valid nbext yaml, right?'), 3423509],
            'missing_key': {str('Main'): True},
            'invalid_type': {
                str('Main'): str('main.js'),
                str('Type'): str('blahblahblah')
            },
            'dummy': {
                str('Main'): str('dummy.js'),
                str('Type'): str('Jupyter Notebook Extension'),
                str('Description'): str('This is a dumb dummy description'),
                str('Compatibility'): str('4.x 5.x'),
            },
        }
        for fname, yaml_obj in test_yamls.items():
            if fname != 'dummy':
                nt.assert_not_is_instance(
                    jupyter_nbextensions_configurator._process_nbextension_spec(yaml_obj),  # noqa: E501
                    dict)
            yaml_path = os.path.join(dodgy_nbext_dir_path, fname + '.yaml')
            with io.open(yaml_path, 'w') as f:
                yaml.dump(yaml_obj, f, default_flow_style=False)

        # a yaml file which shadows an existing nbextension.
        nbdir = os.path.join(
            os.path.dirname(jupyter_nbextensions_configurator.__file__),
            'static')
        nbexts = (
            jupyter_nbextensions_configurator.get_configurable_nbextensions(
                [nbdir], as_dict=True))
        src = random.choice(list(nbexts.values()))['yaml_path']
        dst = os.path.join(
            dodgy_nbext_dir_path, os.path.relpath(src, start=nbdir))
        dst_dir = os.path.dirname(dst)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        shutil.copy(src, dst)



# -*- coding: utf-8 -*-
"""Tests for the main app."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals,
)

import itertools
import json
import logging
import os
from unittest import TestCase

import jupyter_core.paths
import nose.tools as nt
from jupyter_contrib_core.notebook_compat import serverextensions
from jupyter_contrib_core.testing_utils import (
    get_logger, patch_traitlets_app_logs,
)
from jupyter_contrib_core.testing_utils.jupyter_env import patch_jupyter_dirs
from traitlets.config import Config
from traitlets.tests.utils import check_help_all_output, check_help_output

from jupyter_nbextensions_configurator.application import main as main_app
from jupyter_nbextensions_configurator.application import (
    DisableJupyterNbextensionsConfiguratorApp,
    EnableJupyterNbextensionsConfiguratorApp,
    JupyterNbextensionsConfiguratorApp,
)

app_classes = (DisableJupyterNbextensionsConfiguratorApp,
               EnableJupyterNbextensionsConfiguratorApp,
               JupyterNbextensionsConfiguratorApp)


def reset_app_class(app_class):
    """Reset all app traits and clear the instance."""
    if app_class._instance is None:
        return
    for name, traitlet in app_class._instance.traits().items():
        if isinstance(traitlet.this_class, JupyterNbextensionsConfiguratorApp):
            setattr(app_class._instance, name, traitlet.default_value)
    app_class.clear_instance()


class AppTest(TestCase):
    """Tests for the main app."""

    @classmethod
    def setup_class(cls):
        cls.log = cls.log = get_logger(cls.__name__)
        cls.log.handlers = []
        cls.log.propagate = True

    def setUp(self):
        """Set up test fixtures for each test."""
        (jupyter_patches, self.jupyter_dirs,
         remove_jupyter_dirs) = patch_jupyter_dirs()
        for ptch in jupyter_patches:
            ptch.start()
            self.addCleanup(ptch.stop)
        self.addCleanup(remove_jupyter_dirs)

        for klass in app_classes:
            patch_traitlets_app_logs(klass)
            klass.log_level.default_value = logging.DEBUG

    def check_enable(self, argv=None, dirs=None):
        """Check files were enabled in the correct place."""
        if argv is None:
            argv = []
        if dirs is None:
            dirs = {
                'conf': jupyter_core.paths.jupyter_config_dir(),
                'data': jupyter_core.paths.jupyter_data_dir(),
            }
        conf_dir = dirs['conf']

        # do enable
        main_app(argv=['enable'] + argv)

        # list everything that got enabled
        created_files = []
        for root, subdirs, files in os.walk(dirs['conf']):
            created_files.extend([os.path.join(root, f) for f in files])
        nt.assert_true(
            created_files,
            'enable should create files in {}'.format(dirs['conf']))

        # a bit of a hack to allow initializing a new app instance
        for klass in app_classes:
            reset_app_class(klass)

        # do disable
        main_app(argv=['disable'] + argv)
        # check the config directory
        conf_enabled = [
            path for path in created_files
            if path.startswith(conf_dir) and os.path.exists(path)]
        for path in conf_enabled:
            with open(path, 'r') as f:
                conf = Config(json.load(f))
            nbapp = conf.get('NotebookApp', {})
            if 'server_extensions' in nbapp:
                nt.assert_not_in(
                    'jupyter_nbextensions_configurator',
                    nbapp.server_extensions,
                    'conf after disable should empty'
                    'server_extensions list in file {}'.format(path))
            nbservext = nbapp.get('nbserver_extensions', {})
            nt.assert_false(
                {k: v for k, v in nbservext.items() if v},
                'disable command should disable all '
                'nbserver_extensions in file {}'.format(path))

        reset_app_class(DisableJupyterNbextensionsConfiguratorApp)

    def test_00_extra_args(self):
        """Check that app complains about extra args."""
        for subcom in ('enable', 'disable'):
            # sys.exit should be called if extra args specified
            with nt.assert_raises(SystemExit):
                main_app([subcom, 'arbitrary_extension_name'])
            for klass in app_classes:
                reset_app_class(klass)

    def test_01_help_output(self):
        """Check that app help works."""
        app_module = 'jupyter_nbextensions_configurator.application'
        for argv in (['enable'], ['disable']):
            check_help_output(app_module, argv)
            check_help_all_output(app_module, argv)
        # sys.exit should be called if no argv specified
        with nt.assert_raises(SystemExit):
            main_app([])

    def test_02_default_enable(self):
        """Check that enable works correctly using defaults."""
        self.check_enable()

    def test_03_user_enable(self):
        """Check that enable works correctly using --user flag."""
        self.check_enable(argv=['--user'], dirs=self.jupyter_dirs['env_vars'])

    def test_04_sys_prefix_enable(self):
        """Check that enable works correctly using --sys-prefix flag."""
        self.check_enable(
            argv=['--sys-prefix'], dirs=self.jupyter_dirs['sys_prefix'])

    def test_05_system_enable(self):
        """Check that enable works correctly using --system flag."""
        self.check_enable(argv=['--system'], dirs=self.jupyter_dirs['system'])

    def test_06_argument_conflict(self):
        """Check that enable objects to multiple flags."""
        conflicting_flags = ('--user', '--system', '--sys-prefix')
        conflicting_flagsets = []
        for nn in range(2, len(conflicting_flags) + 1):
            conflicting_flagsets.extend(
                itertools.combinations(conflicting_flags, nn))
        for flagset in conflicting_flagsets:
            self.log.info('testing conflicting flagset {}'.format(flagset))
            nt.assert_raises(serverextensions.ArgumentConflict,
                             main_app, ['enable'] + list(flagset))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Setup script for jupyter_nbextensions_configurator."""

from __future__ import print_function

import os
from glob import glob

from setuptools import find_packages, setup


def main():

    setup(
        name='jupyter_nbextensions_configurator',
        description=("jupyter serverextension providing configuration "
                     "interfaces for nbextensions."),
        long_description="""
The jupyter_nbextensions_configurator jupyter server extension provides
graphical user interfaces for configuring which nbextensions are enabled (load
automatically for every notebook), and display their readme files.
In addition, for nbextensions which include an appropriate yaml descriptor
file, the interface also provides controls to configure the nbextensions'
options.
""",
        version='0.4.1',
        author='jcb91, jupyter-contrib developers',
        author_email='joshuacookebarnes@gmail.com',
        url=('https://github.com/'
             'jupyter-contrib/jupyter_nbextensions_configurator'),
        download_url=('https://github.com/'
                      'jupyter-contrib/jupyter_nbextensions_configurator/'
                      'tarball/0.4.1'),
        keywords=['Jupyter', 'notebook'],
        license='BSD 3-clause',
        platforms=['any'],
        packages=find_packages('src'),
        package_dir={'': 'src'},
        include_package_data=True,
        py_modules=[
            os.path.splitext(os.path.basename(path))[0]
            for path in glob('src/*.py')
        ],
        install_requires=[
            'jupyter_contrib_core >=0.3.3',
            'jupyter_core',
            'notebook >=4.0',
            'pyyaml',
            'tornado',
            'traitlets',
        ],
        extras_require={
            'test': [
                'jupyter_contrib_core[testing_utils]',
                'nose',
                'requests',
                'selenium',
            ],
            'test:python_version == "2.7"': [
                'mock',
            ],
        },
        # we can't be zip safe as we require templates etc to be accessible to
        # jupyter server
        zip_safe=False,
        entry_points={
            'console_scripts': [
                'jupyter-nbextensions_configurator = jupyter_nbextensions_configurator.application:main',  # noqa
            ],
        },
        scripts=[os.path.join('scripts', p) for p in [
            'jupyter-nbextensions_configurator',
        ]],
        classifiers=[
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Natural Language :: English',
            'Operating System :: OS Independent',
            'Programming Language :: JavaScript',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Topic :: Utilities',
        ],
    )


if __name__ == '__main__':
    main()
